import pandas as pd
import numpy as np
import pystan
import platform
import pickle
import scipy.stats as ss
import matplotlib.pyplot as plt
import seaborn as sb

in_flnm = input('Input file(*.xlsx) = ')       #   Excel file    
out_flnm = 'Output.txt'
out_f = open(out_flnm, 'w')
out_f.write('Input data file = ' + in_flnm + '\n')
#
#        Read in the Excel file
#
fl_xlsx = pd.ExcelFile(in_flnm)
data = pd.read_excel(fl_xlsx, header = None)
data = data.values
K = data[0][1]
print('K =', K)
print(data[:10])
out_f.write('\ndata...\n')
for v in data:
    out_f.write(v.__str__() + '\n')
item_names = data[1]
print(item_names)
ck_items = data[2]
sel_items = ck_items == 1
print(sel_items)
item_names = item_names[sel_items]
print('Selected items...')
print(item_names)
out_f.write('\n\nThe selected item names...' + item_names.__str__() + '\n')

data = np.array(data)
X = data[3:].T[1:].T
X = X[:, sel_items[1:]]
print(X[:3])
#
#        MCMC sampling by Stan
#
sm = pystan.StanModel(file = 'calc_omega.stan')

X = X.astype(dtype = int)
fit = sm.sampling(data = {'n_items':len(X[0]), 'n_persons':len(X), 'K':K, 'X': X},
                  pars = ['mu', 'lmbd', 'sgm', 'F'],
                             n_jobs = 1 if platform.system() == 'Windows' else -1)

#
#           Results of MCMC sampling
#
print(fit)
out_f.write('\nfit...\n')
out_f.write(fit.__str__())
#
#      Calculation on parameter mu
#
print('mu =\n', fit['mu'[:10]])
out_f.write('\nMu (Q1, Med, Q3)...\n')
for j in range(len(item_names)):
    out_f.write('    ' + item_names[j] + ':  ')
    mu_q1, mu_med, mu_q3 = np.percentile(fit['mu'].T[j], [25, 50, 75])
    out_f.write('{0:.3f}   {1:.3f}   {2:.3f}\n'.format(mu_q1, mu_med, mu_q3))
#
#       Calculation on parameter lambda
#
print('lmbd =\n', fit['lmbd'][:10])
print('Lambda =\n', fit['lmbd'[:10]])
out_f.write('\nLambda (Q1, Med, Q3)...\n')
for j in range(len(item_names)):
    out_f.write('    ' + item_names[j] + ':  ')
    lmbd_q1, lmbd_med, lmbd_q3 = np.percentile(fit['lmbd'].T[j], [25, 50, 75])
    out_f.write('{0:.3f}   {1:.3f}   {2:.3f}\n'.format(lmbd_q1, lmbd_med, lmbd_q3))
#
#         Calculation on parameter sigma
#
print('Sigma =\n', fit['sgm'[:10]])
out_f.write('\nSigma (Q1, Med, Q3)...\n')
for j in range(len(item_names)):
    out_f.write('    ' + item_names[j] + ':  ')
    sgm_q1, sgm_med, sgm_q3 = np.percentile(fit['sgm'].T[j], [25, 50, 75])
    out_f.write('{0:.3f}   {1:.3f}   {2:.3f}\n'.format(sgm_q1, sgm_med, sgm_q3))
#
#         Calculation on omega
#
sum_lmbd = fit['lmbd'].sum(axis=1)
sum_sqr_sgm = (fit['sgm']**2).sum(axis=1)
omegas = sum_lmbd**2 / (sum_lmbd**2 + sum_sqr_sgm)

def calcMAPEst(samples, a = 0.05, n_points = 1000):
    """         Estimation of the MAP estimate by kde
         samples:    Input data for which MAP estimate is calculated.
         a:          Each side of a/2 of samples is ignored, Default value is 0.05
         n_points:   number of points used in kde.  Default value is 1000
    """
    import numpy as np
    import scipy.stats as ss

    Lp, Up = np.percentile(samples, [100 * a/2, 100 * (1 - a/2)])
    coord = np.linspace(Lp, Up, n_points)
    est_pdf = ss.gaussian_kde(samples).pdf(coord) 
    map_idx = np.argmax(est_pdf)
    MAP_Est = coord[map_idx]      
    return MAP_Est, est_pdf[map_idx]    #   MAP estimate,  the value at MAP estimate


MAP = calcMAPEst(omegas)[0]
print('MAP =', MAP)
Q1, Med, Q3 = np.percentile(omegas, [25, 50, 75])
print(Q1, Med, Q3)
out_f.write('\nomega (MAP / Q1, Med, Q3)...\n')
out_f.write('    {0:.3f}\n'.format(MAP))
out_f.write('    {0:.3f}   {1:.3f}   {2:.3f}\n'.format(Q1, Med, Q3))

sb.kdeplot(omegas, color = 'b')
plt.title('MAP = {0:.2f}\nQ1 = {1:.2f}  Med = {2:.2f}  Q3 = {3:.2f}'.format(MAP, Q1, Med, Q3))
plt.xlabel('$\omega$', fontsize = 14)
plt.savefig('FigOmega.png')
plt.show()
#
#          Calculation on Pearson correlation coefficient
#            between factor F and the sum score of Xjs
#
SumX = X.sum(axis=1)
Cors = []
for Fvalues in fit['F']:
    r = ss.pearsonr(SumX, Fvalues)[0]
    Cors.append(r)
    
r_map = calcMAPEst(Cors)[0]
r_q1, r_med, r_q3 = np.percentile(Cors, [25, 50, 75])

out_f.write('\nCorrelation between F and X (MAP / Q1, Med, Q3)...\n')
out_f.write('    {0:.3f}\n'.format(r_map))
out_f.write('    {0:.3f}   {1:.3f}   {2:.3f}\n'.format(r_q1, r_med, r_q3))
    
sb.kdeplot(Cors)
plt.title('MAP = {0:.2f}\nQ1 = {1:.2f}  Med = {2:.2f}  Q3 = {3:.2f}'.
         format(r_map, r_q1, r_med, r_q3), fontsize = 14)
plt.xlabel('r', fontsize = 16)
plt.savefig('FigCors.png')
plt.show()
print('FigOmega.png was saved.')
print('FigCors,png was saved.')
out_f.close()
print(out_flnm + ' was saved.')
